/**
* @file Douglas Crockford's monad library as UMD/ES6 module
* @version 1.0.6
* Based on Douglas Crockford's Public Domain monad.js library.
* This is an almost verbatim copy of Douglas Crockford's work, but
* made easy for use in browser/node.js, or as an ES6 module.
*
* As github does not provide a Public Domain option as one of the
* supported licenses, this work uses the liberal MIT license instead.
* @copyright (c) 2017 Greg Borota.
* @license MIT
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*/

// https://github.com/douglascrockford/monad
// https://gist.github.com/newswim/4668aef8a1f1bc0dabe8
// https://www.youtube.com/watch?v=dkZFtimgAcM
// https://www.youtube.com/watch?v=ZhuHCtR3xq8
// https://blogs.msdn.microsoft.com/wesdyer/2008/01/10/the-marvels-of-monads/
/** @module monadJs */

/**
 * The `makeMonad` function is a macroid that produces monad constructor
 * functions. It can take an optional `modifier` function, which is a function
 * that is allowed to modify new monads at the end of the construction processes.
 *
 * A monad constructor (sometimes called `unit` or `return` in some mythologies)
 * comes with three methods, `lift`, `liftValue`, and `method`, all of which can
 * add methods and properties to the monad's prototype.
 *
 * A monad has a `bind` method that takes a function that receives a value and
 * is usually expected to return a monad.
 * ```
 *    const identity = makeMonad();
 *    const monad = identity("Hello world.");
 *    monad.bind(alert);
 *
 *    const ajax = makeMonad()
 *      .lift('alert', alert);
 *    const monad = ajax("Hello world.");
 *    monad.alert();
 *
 *    const maybe = makeMonad(function (monad, value) {
 *        if (value === null || value === undefined) {
 *            monad.isNull = true;
 *            monad.bind = function () {
 *                return monad;
 *            };
 *            return null;
 *        }
 *        return value;
 *    });
 *    const monad = maybe(null);
 *    monad.bind(alert);    // Nothing happens.
 * ```
 * @name makeMonad
 * @static
 */
function makeMonad(modifier) {
  // Each unit constructor has a monad prototype. The prototype will contain an
  // isMonad property for classification, as well as all inheritable methods.

  var prototype = Object.create(null);
  prototype.isMonad = true;

  // Each call to makeMonad will produce a new unit constructor function.

  function unit(value) {
    // Construct a new monad.

    var monad = Object.create(prototype);

    // In some mythologies 'bind' is called 'pipe' or '>>='.
    // The bind method will deliver the unit's value parameter to a function.

    monad.bind = function bind(func, args) {
      // bind takes a function and an optional array of arguments. It calls that
      // function passing the monad's value and bind's optional array of args.

      return func.apply(void 0, [ value ].concat( args ));
    };

    // If makeMonad's modifier parameter is a function, then call it, passing the monad
    // and the value.

    if (typeof modifier === 'function') {
      value = modifier(monad, value);
    }

    // Return the shiny new monad.

    return monad;
  }
  unit.method = function method(name, func) {
    // Add a method to the prototype.

    prototype[name] = func;
    return unit;
  };
  unit.liftValue = function liftValue(name, func) {
    // Add a method to the prototype that calls bind with the func. This can be
    // used for ajax methods that return values other than monads.

    prototype[name] = function liftedValue() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return this.bind(func, args);
    };
    return unit;
  };
  unit.lift = function lift(name, func) {
    // Add a method to the prototype that calls bind with the func. If the value
    // returned by the func is not a monad, then make a monad.

    prototype[name] = function lifted() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      var result = this.bind(func, args);
      return result && result.isMonad === true ? result : unit(result);
    };
    return unit;
  };
  return unit;
}

/** @module monadJs */
/**
 * The Identity Monad
 * ```
 *  const monad = identity("Hello world.");
 *  monad.bind(alert);
 * ```
 * @name identity
 * @static
 */
var identity = makeMonad();

/** @module monadJs */
/**
 * Maybe monad is used to guard against Null Pointer Exceptions.
 * ```
 *  const monad = maybe(null);
 *  monad.bind(alert);    // Nothing happens.
 * ```
 * @name maybe
 * @static
 */
var maybe = makeMonad(function (monad, value) {
  if (value === null || value === undefined) {
    monad.isNull = true;
    monad.bind = function bind() {
      return monad;
    };
    return null;
  }
  monad.isNull = false;
  return value;
});

/** @module monadJs */

/**
 * Create a `vow` object, used to handle promises
 *
 * NOTE: ES6 has native support for Promises. This implementation could still be
 * used to create promises in ES5 though.
 * @const
 * @name vow
 * @static
 */
var VOW = (function deferrer() {
// The VOW object contains a .make function that is used to make vows.
// It may also contain other useful functions.
// In some mythologies, 'VOW' is called 'deferrer'.

  function enlighten(queue, fate) {
  // enlighten is a helper function of herald and .when. It schedules the
  // processing of all of the resolution functions in either the keepers queue
  // or the breakers queue in later turns with the promise's fate.

    queue.forEach(function (func) {
      setImmediate(func, fate);
    });
  }

  return {
    make: function make() {
    // The make function makes new vows. A vow contains a promise object and the
    // two resolution functions (break and keep) that determine the fate of the
    // promise.

      var breakers = []; // .when's broken queue
      var fate; // The promise's ultimate value
      var keepers = []; // .when's kept queue
      var status = 'pending'; // 'broken'; 'kept'; or 'pending'

      function enqueue(
        resolution, // 'keep' or 'break'
        func, // A function that was registered with .when
        vow // A vow that provides the resolution functions
      ) {
        // enqueue is a helper function used by .when. It will append a function to
        // either the keepers queue or the breakers queue.

        var queue = (resolution === 'keep')
          ? keepers
          : breakers;
        queue[queue.length] = (typeof func !== 'function')

          // If func is not a function, push the resolver so that the value passes to
          // the next cascaded .when.

          ? vow[resolution]

          // If the func is a function, push a function that calls func with a value.
          // The result can be a promise, or not a promise, or an exception.

          : function queued(value) {
            try {
              var result = func(value);

              // If the result is a promise, then register our resolver with that promise.

              if (result && result.isPromise === true) {
                result.when(vow.keep, vow.break);

              // But if it is not a promise, then use the result to resolve our promise.
              } else {
                vow.keep(result);
              }

            // But if func throws an exception, then break our promise.
            } catch (e) {
              vow.break(e);
            }
          };
      }

      function herald(state, value, queue) {
      // The herald function is a helper function of break and keep.
      // It seals the promise's fate, updates its status, enlightens
      // one of the queues, and empties both queues.

        if (status !== 'pending') {
          throw new Error('overpromise');
        }
        fate = value;
        status = state;
        enlighten(queue, fate);
        keepers.length = 0;
        breakers.length = 0;
      }

      // Construct and return the vow object.

      return {
        break: function breakPromise(value) {
        // The break method breaks the promise.

          herald('broken', value, breakers);
        },
        keep: function keepPromise(value) {
        // The keep method keeps the promise.

          herald('kept', value, keepers);
        },
        promise: {

          // The promise is an object with a .when method.

          isPromise: true,

          // The .when method is the promise monad's bind. The .when method can take two
          // optional functions. One of those functions may be called, depending on the
          // promise's resolution. Both could be called if the the kept function throws.

          when: function makeVow(kept, broken) {
          // Make a new vow. Return the new promise.

            var vow = make();
            switch (status) {
            // If this promise is still pending, then enqueue both kept and broken.

              case 'pending':
                enqueue('keep', kept, vow);
                enqueue('break', broken, vow);
                break;

              // If the promise has already been kept, then enqueue only the kept function,
              // and enlighten it.

              case 'kept':
                enqueue('keep', kept, vow);
                enlighten(keepers, fate);
                break;

              default:
                break;

              // If the promise has already been broken, then enqueue only the broken
              // function, and enlighten it.

              case 'broken':
                enqueue('break', broken, vow);
                enlighten(breakers, fate);
                break;
            }
            return vow.promise;
          }
        }
      };
    },
    every: function every(array) {
    // The every function takes an array of promises and returns a promise that
    // will deliver an array of results only if every promise is kept.

      var remaining = array.length;
      var results = [];
      var vow = VOW.make();

      if (!remaining) {
        vow.break(array);
      } else {
        array.forEach(function (promise, i) {
          promise.when(function (value) {
            results[i] = value;
            remaining -= 1;
            if (remaining === 0) {
              vow.keep(results);
            }
          }, function (reason) {
            remaining = NaN;
            vow.break(reason);
          });
        });
      }
      return vow.promise;
    },
    first: function first(array) {
    // The first function takes an array of promises and returns a promise to
    // deliver the first observed kept promise, or a broken promise if all of
    // the promises are broken.

      var found = false;
      var remaining = array.length;
      var vow = VOW.make();

      function check() {
        remaining -= 1;
        if (remaining === 0 && !found) {
          vow.break();
        }
      }

      if (remaining === 0) {
        vow.break(array);
      } else {
        array.forEach(function (promise) {
          promise.when(function (value) {
            if (!found) {
              found = true;
              vow.keep(value);
            }
            check();
          }, check);
        });
      }
      return vow.promise;
    },
    any: function any(array) {
    // The any function takes an array of promises and returns a promise that
    // will deliver a possibly sparse array of results of any kept promises.
    // The result will contain an undefined element for each broken promise.

      var remaining = array.length;
      var results = [];
      var vow = VOW.make();

      function check() {
        remaining -= 1;
        if (remaining === 0) {
          vow.keep(results);
        }
      }

      if (!remaining) {
        vow.keep(results);
      } else {
        array.forEach(function (promise, i) {
          promise.when(function (value) {
            results[i] = value;
            check();
          }, check);
        });
      }
      return vow.promise;
    },
    kept: function keptPromise(value) {
    // Returns a new kept promise.

      var vow = VOW.make();
      vow.keep(value);
      return vow.promise;
    },
    broken: function brokenPromise(reason) {
    // Returns a new broken promise.

      var vow = VOW.make();
      vow.break(reason);
      return vow.promise;
    }
  };
}());

export { makeMonad, identity, maybe, VOW as vow };
